import pytest
import pyeio as io


# from pyeio.core.exceptions import (
#     MissingFileExtensionError,
#     UnsupportedFileExtensionError,
#     IncorrectFileExtensionError,
# )




# def test_parse_types():
#     with raises(TypeError):
#         _ = json.parse(data={"asdf", "1234"})
